package com.sun.jndi.ldap;

import javax.naming.directory.ModificationItem;
import javax.naming.NamingEnumeration;
import javax.naming.directory.Attribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.CompositeName;
import javax.naming.NameNotFoundException;
import javax.naming.directory.SchemaViolationException;
import javax.naming.Name;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import javax.naming.directory.Attributes;
import java.util.Hashtable;
import com.sun.jndi.toolkit.dir.HierMemDirCtx;

final class LdapSchemaCtx extends HierMemDirCtx
{
    private static final boolean debug = false;
    private static final int LEAF = 0;
    private static final int SCHEMA_ROOT = 1;
    static final int OBJECTCLASS_ROOT = 2;
    static final int ATTRIBUTE_ROOT = 3;
    static final int SYNTAX_ROOT = 4;
    static final int MATCHRULE_ROOT = 5;
    static final int OBJECTCLASS = 6;
    static final int ATTRIBUTE = 7;
    static final int SYNTAX = 8;
    static final int MATCHRULE = 9;
    private LdapSchemaCtx.SchemaInfo info;
    private boolean setupMode;
    private int objectType;
    
    static DirContext createSchemaTree(Hashtable<String, Object> hashtable, String s, LdapCtx ldapCtx, Attributes attributes, boolean b) throws NamingException {
        try {
            LdapSchemaCtx ldapSchemaCtx = new LdapSchemaCtx(1, hashtable, new LdapSchemaCtx.SchemaInfo(s, ldapCtx, new LdapSchemaParser(b)));
            LdapSchemaParser.LDAP2JNDISchema(attributes, ldapSchemaCtx);
            return ldapSchemaCtx;
        }
        catch (NamingException ex) {
            ldapCtx.close();
            throw ex;
        }
    }
    
    private LdapSchemaCtx(int objectType, Hashtable<String, Object> environment, LdapSchemaCtx.SchemaInfo info) {
        super(environment, true);
        this.info = null;
        this.setupMode = true;
        this.objectType = objectType;
        this.info = info;
    }
    
    @Override
    public void close() throws NamingException {
        this.info.close();
    }
    
    @Override
    public final void bind(Name name, Object o, Attributes attrs) throws NamingException {
        if (!this.setupMode) {
            if (o != null) {
                throw new IllegalArgumentException("obj must be null");
            }
            this.addServerSchema(attrs);
        }
        LdapSchemaCtx ldapSchemaCtx = (LdapSchemaCtx)super.doCreateSubcontext(name, attrs);
    }
    
    @Override
    protected final void doBind(Name name, Object obj, Attributes attrs, boolean b) throws NamingException {
        if (!this.setupMode) {
            throw new SchemaViolationException("Cannot bind arbitrary object; use createSubcontext()");
        }
        super.doBind(name, obj, attrs, false);
        return;
    }
    
    @Override
    public final void rebind(Name name, Object o, Attributes attributes) throws NamingException {
        try {
            this.doLookup(name, false);
            throw new SchemaViolationException("Cannot replace existing schema object");
        }
        catch (NameNotFoundException ex) {
            this.bind(name, o, attributes);
            return;
        }
    }
    
    @Override
    protected final void doRebind(Name name, Object obj, Attributes attrs, boolean b) throws NamingException {
        if (!this.setupMode) {
            throw new SchemaViolationException("Cannot bind arbitrary object; use createSubcontext()");
        }
        super.doRebind(name, obj, attrs, false);
        return;
    }
    
    @Override
    protected final void doUnbind(Name name) throws NamingException {
        if (!this.setupMode) {
            try {
                this.deleteServerSchema(((LdapSchemaCtx)this.doLookup(name, false)).attrs);
            }
            catch (NameNotFoundException ex) {
                return;
            }
        }
        super.doUnbind(name);
    }
    
    @Override
    protected final void doRename(Name oldname, Name newname) throws NamingException {
        if (!this.setupMode) {
            throw new SchemaViolationException("Cannot rename a schema object");
        }
        super.doRename(oldname, newname);
        return;
    }
    
    @Override
    protected final void doDestroySubcontext(Name name) throws NamingException {
        if (!this.setupMode) {
            try {
                this.deleteServerSchema(((LdapSchemaCtx)this.doLookup(name, false)).attrs);
            }
            catch (NameNotFoundException ex) {
                return;
            }
        }
        super.doDestroySubcontext(name);
    }
    
    final LdapSchemaCtx setup(int objectType, String n, Attributes attrs) throws NamingException {
        try {
            this.setupMode = true;
            LdapSchemaCtx ldapSchemaCtx = (LdapSchemaCtx)super.doCreateSubcontext(new CompositeName(n), attrs);
            ldapSchemaCtx.objectType = objectType;
            ldapSchemaCtx.setupMode = false;
            return ldapSchemaCtx;
        }
        finally {
            this.setupMode = false;
        }
    }
    
    @Override
    protected final DirContext doCreateSubcontext(Name name, Attributes attrs) throws NamingException {
        if (attrs == null || attrs.size() == 0) {
            throw new SchemaViolationException("Must supply attributes describing schema");
        }
        if (!this.setupMode) {
            this.addServerSchema(attrs);
        }
        return (DirContext)super.doCreateSubcontext(name, attrs);
    }
    
    private static final Attributes deepClone(Attributes attributes) throws NamingException {
        BasicAttributes basicAttributes = new BasicAttributes(true);
        NamingEnumeration<Object> all = attributes.getAll();
        while (all.hasMore()) {
            basicAttributes.put((Attribute)((Attribute)all.next()).clone());
        }
        return basicAttributes;
    }
    
    @Override
    protected final void doModifyAttributes(ModificationItem[] array) throws NamingException {
        if (this.setupMode) {
            super.doModifyAttributes(array);
        }
        else {
            Attributes deepClone = deepClone(this.attrs);
            HierMemDirCtx.applyMods(array, deepClone);
            this.modifyServerSchema(this.attrs, deepClone);
            this.attrs = deepClone;
        }
    }
    
    @Override
    protected final HierMemDirCtx createNewCtx() {
        return new LdapSchemaCtx(0, this.myEnv, this.info);
    }
    
    private final void addServerSchema(Attributes attributes) throws NamingException {
        Attribute attribute;
        switch (this.objectType) {
            case 2: {
                attribute = this.info.parser.stringifyObjDesc(attributes);
                break;
            }
            case 3: {
                attribute = this.info.parser.stringifyAttrDesc(attributes);
                break;
            }
            case 4: {
                attribute = this.info.parser.stringifySyntaxDesc(attributes);
                break;
            }
            case 5: {
                attribute = this.info.parser.stringifyMatchRuleDesc(attributes);
                break;
            }
            case 1: {
                throw new SchemaViolationException("Cannot create new entry under schema root");
            }
            default: {
                throw new SchemaViolationException("Cannot create child of schema object");
            }
        }
        BasicAttributes o = new BasicAttributes(true);
        o.put(attribute);
        this.info.modifyAttributes(this.myEnv, 1, o);
    }
    
    private final void deleteServerSchema(Attributes attributes) throws NamingException {
        Attribute attr;
        switch (this.objectType) {
            case 2: {
                attr = this.info.parser.stringifyObjDesc(attributes);
                break;
            }
            case 3: {
                attr = this.info.parser.stringifyAttrDesc(attributes);
                break;
            }
            case 4: {
                attr = this.info.parser.stringifySyntaxDesc(attributes);
                break;
            }
            case 5: {
                attr = this.info.parser.stringifyMatchRuleDesc(attributes);
                break;
            }
            case 1: {
                throw new SchemaViolationException("Cannot delete schema root");
            }
            default: {
                throw new SchemaViolationException("Cannot delete child of schema object");
            }
        }
        this.info.modifyAttributes(this.myEnv, new ModificationItem[] { new ModificationItem(3, attr) });
    }
    
    private final void modifyServerSchema(Attributes attributes, Attributes attributes2) throws NamingException {
        Attribute attr;
        Attribute attr2;
        switch (this.objectType) {
            case 6: {
                attr = this.info.parser.stringifyObjDesc(attributes);
                attr2 = this.info.parser.stringifyObjDesc(attributes2);
                break;
            }
            case 7: {
                attr = this.info.parser.stringifyAttrDesc(attributes);
                attr2 = this.info.parser.stringifyAttrDesc(attributes2);
                break;
            }
            case 8: {
                attr = this.info.parser.stringifySyntaxDesc(attributes);
                attr2 = this.info.parser.stringifySyntaxDesc(attributes2);
                break;
            }
            case 9: {
                attr = this.info.parser.stringifyMatchRuleDesc(attributes);
                attr2 = this.info.parser.stringifyMatchRuleDesc(attributes2);
                break;
            }
            default: {
                throw new SchemaViolationException("Cannot modify schema root");
            }
        }
        this.info.modifyAttributes(this.myEnv, new ModificationItem[] { new ModificationItem(3, attr), new ModificationItem(1, attr2) });
    }
    
    private static final class SchemaInfo
    {
        private LdapCtx schemaEntry;
        private String schemaEntryName;
        LdapSchemaParser parser;
        private String host;
        private int port;
        private boolean hasLdapsScheme;
        
        SchemaInfo(String schemaEntryName, LdapCtx schemaEntry, LdapSchemaParser parser) {
            this.schemaEntryName = schemaEntryName;
            this.schemaEntry = schemaEntry;
            this.parser = parser;
            this.port = schemaEntry.port_number;
            this.host = schemaEntry.hostname;
            this.hasLdapsScheme = schemaEntry.hasLdapsScheme;
        }
        
        synchronized void close() throws NamingException {
            if (this.schemaEntry != null) {
                this.schemaEntry.close();
                this.schemaEntry = null;
            }
        }
        
        private LdapCtx reopenEntry(Hashtable<?, ?> hashtable) throws NamingException {
            return new LdapCtx(this.schemaEntryName, this.host, this.port, hashtable, this.hasLdapsScheme);
        }
        
        synchronized void modifyAttributes(Hashtable<?, ?> hashtable, ModificationItem[] mods) throws NamingException {
            if (this.schemaEntry == null) {
                this.schemaEntry = this.reopenEntry(hashtable);
            }
            this.schemaEntry.modifyAttributes("", mods);
        }
        
        synchronized void modifyAttributes(Hashtable<?, ?> hashtable, int mod_op, Attributes attrs) throws NamingException {
            if (this.schemaEntry == null) {
                this.schemaEntry = this.reopenEntry(hashtable);
            }
            this.schemaEntry.modifyAttributes("", mod_op, attrs);
        }
    }
}
